import logging
from pynput.keyboard import Key
from pynput.mouse import Button
# from pynput._util.xorg import X11Error
from sortedcontainers import SortedDict

log = logging.getLogger(__name__)

class Mapping(object):
    def __init__(self, m, k):
        self.m = m
        self.k = k

    def action(self, key, release=True, reset_modifiers=True):
        """Returns action function to be called on a touch"""
        def action():
            # FIXME: upstream pynput bug, caps_lock raises exception if mapped to mod key
            try:
                self.k.press(key)
                if release:
                    self.k.release(key)
                if reset_modifiers:
                    self.k.release(Key.shift)
                    self.k.release(Key.ctrl)
                    self.k.release(Key.alt)
                    self.k.release(Key.shift_r)
                    self.k.release(Key.ctrl_r)

            # except X11Error:
            except Exception:
                log.error(f"X11Error for key {key}")

        return action

    def call_action(self, x, y):
        """Look up coordinate in table and call action"""
        regions = self.regions()
        row_index = regions.bisect(y)
        # FIXME: should be a cleaner way to limit row_index to valid indices
        row_index = row_index if row_index < len(regions) else len(regions) - 1
        row = regions[regions.keys()[row_index]]
        col_index = row.bisect(x)
        # FIXME: should be a cleaner way to limit col_index to valid indices
        col_index = col_index if col_index < len(row) else len(row) - 1
        # call function
        row[row.keys()[col_index]]()


class KeyboardMapping(Mapping):
    def regions(self):
        return SortedDict({
            260: SortedDict({
                1499: lambda: self.m.click(Button.left),
                1872: lambda: self.m.scroll(0, 1)
            }),
            519: SortedDict({
                1499: lambda: self.m.click(Button.left),
                1872: lambda: self.m.scroll(0, -1)
            }),
            779: SortedDict({
                1499: lambda: self.m.click(Button.left),
                1872: lambda: self.m.click(Button.right)
            }),
            903: SortedDict({
                126: self.action('`'),
                251: self.action('1'),
                376: self.action('2'),
                500: self.action('3'),
                625: self.action('4'),
                750: self.action('5'),
                874: self.action('6'),
                1000: self.action('7'),
                1125: self.action('8'),
                1249: self.action('9'),
                1374: self.action('0'),
                1499: self.action('-'),
                1623: self.action('='),
                1872: self.action(Key.backspace)
            }),
            1028: SortedDict({
                188: self.action(Key.tab),
                314: self.action('q'),
                438: self.action('w'),
                563: self.action('e'),
                688: self.action('r'),
                813: self.action('t'),
                937: self.action('y'),
                1062: self.action('u'),
                1187: self.action('i'),
                1312: self.action('o'),
                1436: self.action('p'),
                1561: self.action('{'),
                1686: self.action('}'),
                1872: self.action('|')
            }),
            1153: SortedDict({
                220: self.action(Key.caps_lock, release=False),
                345: self.action('a'),
                470: self.action('s'),
                594: self.action('d'),
                719: self.action('f'),
                844: self.action('g'),
                968: self.action('h'),
                1093: self.action('j'),
                1218: self.action('k'),
                1343: self.action('l'),
                1468: self.action(';'),
                1598: self.action('\''),
                1872: self.action(Key.enter)
            }),
            1278: SortedDict({
                282: self.action(Key.shift, release=False, reset_modifiers=False),
                407: self.action('z'),
                531: self.action('x'),
                657: self.action('c'),
                781: self.action('v'),
                906: self.action('b'),
                1031: self.action('n'),
                1155: self.action('m'),
                1280: self.action(','),
                1405: self.action('.'),
                1530: self.action('/'),
                1872: self.action(Key.shift_r, release=False, reset_modifiers=False)
            }),
            1404: SortedDict({
                188: self.action(Key.ctrl, release=False, reset_modifiers=False),
                313: self.action(Key.cmd),
                500: self.action(Key.alt, release=False, reset_modifiers=False),
                1250: self.action(Key.space),
                1374: self.action(Key.esc),
                1499: self.action(Key.left),
                1623: self.action(Key.down),
                1747: self.action(Key.up),
                1872: self.action(Key.right)
            }),
        })


class ExampleMapping(Mapping):
    """Example mapping

           +--- x
           |
           |  +-------------------+-+   ---
           y  |         |         |o|    |
              |    a    |    b    | |    |
              |---------+---------|o|  1404
              |         |         | |    |
              |    c    |    d    |o|    |
              +-------------------+-+   ---

              |-------1872--------|

    Coordinates defined in terms of image size: 1872 x 1404
    """
    def regions(self):
        return SortedDict({
            1404 // 2: SortedDict({
                1872 // 2: lambda: self.k.type('a'),
                1872: lambda: self.k.type('b')
            }),
            1404: SortedDict({
                1872 // 2: lambda: self.k.type('c'),
                1872: lambda: self.k.type('d')
            })
        })



from setuptools import setup
from remarkable_keyboard import version

setup(
    name='remarkable-keyboard',
    version=version.__version__,
    packages=['remarkable_keyboard'],
    author="Evan Widloski",
    author_email="evan@evanw.org",
    description="use reMarkable as a wireless mouse and keyboard",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license="GPLv3",
    keywords="remarkable tablet evdev mouse keyboard",
    url="https://github.com/evidlo/remarkable_keyboard",
    entry_points={
        'console_scripts': [
            'remarkable-keyboard = remarkable_keyboard.remarkable_keyboard:main',
            'rekeyboard = remarkable_keyboard.remarkable_keyboard:main'
        ]
    },
    install_requires=[
        'pycryptodome',
        'paramiko',
        'screeninfo',
        'pynput',
        'remarkable-mouse>=5',
        'sortedcontainers'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
    ]
)



#!/bin/env python
# Evan Widloski - 2019-02-23
# Use reMarkable as mouse input

import argparse
import logging
import os
import sys
import struct
from getpass import getpass
import paramiko
from screeninfo import get_monitors
from pynput.mouse import Button, Controller as MouseController
from pynput.keyboard import Key, Controller as KeyboardController
from remarkable_mouse.remarkable_mouse import open_rm_inputs
from remarkable_keyboard.mappings import KeyboardMapping


# evdev codes
e_type_abs = 3
e_code_finger_xpos = 53
e_code_finger_ypos = 54
e_code_finger_pressure = 58
e_code_finger_touch = 57
e_value_finger_up = -1

# resolution of reMarkable touchscreen
finger_width = 1023
finger_height = 767

# resolution of reMarkable display
image_width = 1872
image_height = 1404


log = logging.getLogger(__name__)

# remap finger coordinates to standard basis
def map_finger(x, y, orig_width, orig_height, map_width, map_height):
    """Map touchscreen coordinates to new axes


    reMarkable is assumed to be laying horizontally on left edge.

    Finger coordinate axes      Mapped coordinate axes
     x ---+                     +--- x
          |                     |
          |                     |
          y                     y


    Coordinates are mapped so that the entire desktop screen fits onto reMarkable


           Monitor wider than reMarkable    Monitor narrower than remarkable

       ---    +-------------------+-+          +-------------------+-+
        |     |                   |o|          |   MMMMMMMMMMMMM   |o|
        |     |MMMMMMMMMMMMMMMMMMM| |          |   MMMMMMMMMMMMM   | |
      height  |MMMMMMMMMMMMMMMMMMM|o|          |   MMMMMMMMMMMMM   |o|
        |     |MMMMMMMMMMMMMMMMMMM| |          |   MMMMMMMMMMMMM   | |
        |     |                   |o|          |   MMMMMMMMMMMMM   |o|
       ---    +-------------------+-+          +-------------------+-+

              |----- width -------|

                             M = monitor mapped area
    """

    x = orig_width - x

    ratio_width, ratio_height = map_width / orig_width, map_height / orig_height
    scaling = ratio_width if ratio_width > ratio_height else ratio_height

    return (
        scaling * (x - (orig_width - map_width / scaling) / 2),
        scaling * (y - (orig_height - map_height / scaling) / 2)
    )


def read_tablet(args):
    """Loop forever and map evdev events to mouse"""

    last_x = last_y = None
    start_x = start_y = x = y = 0
    moves = 0

    monitor = get_monitors()[0]
    log.debug('Chose monitor: {}'.format(monitor))

    touch = open_rm_inputs(address=args.address, key=args.key, password=args.password)['touch']

    mouse = MouseController()
    keyboard = KeyboardController()
    # table of image coordinates and actions
    mapping = KeyboardMapping(mouse, keyboard)

    while True:
        _, _, e_type, e_code, e_value = struct.unpack('2IHHi', touch.read(16))

        if e_type == e_type_abs:

            # handle x direction (rotated)
            if e_code == e_code_finger_ypos:
                log.debug(e_value)
                x = e_value
                monitor_x, monitor_y = map_finger(
                    x, y,
                    finger_width, finger_height,
                    monitor.width, monitor.height
                )
                # use relative movement instead of absolute position
                # (like a trackpad)
                if last_x is not None:
                    mouse.move(monitor_x - last_x, 0)
                    moves += 1
                else:
                    start_x = x
                last_x = monitor_x

            # handle y direction (rotated)
            elif e_code == e_code_finger_xpos:
                log.debug('\t{}'.format(e_value))
                y = e_value
                monitor_x, monitor_y = map_finger(
                    x, y,
                    finger_width, finger_height,
                    monitor.width, monitor.height
                )
                # use relative movement instead of absolute position
                # (like a trackpad)
                if last_y is not None:
                    mouse.move(0, monitor_y - last_y)
                    moves += 1
                else:
                    start_y = y
                last_y = monitor_y

            # handle pressure
            elif e_code == e_code_finger_pressure:
                log.debug('\t\t{}'.format(e_value))
                pass

            # handle finger lift
            elif e_code == e_code_finger_touch:
                if e_value == e_value_finger_up:
                    log.debug('\t\t\tfinger up')
                    last_x = None
                    last_y = None
                    log.debug('start position:{}  {}'.format(start_x, start_y))
                    # FIXME: I have no freaking idea why, but the keyboard lags slightly unless
                    # we print something to console here
                    print(' \r', end='')
                    # if the mouse hasn't moved much since the last finger lift, assume it was a touch
                    if moves < 8:
                        log.debug('touch, moves:{}'.format(moves))
                        image_x, image_y = map_finger(
                            start_x, start_y,
                            finger_width, finger_height,
                            image_width, image_height
                        )
                        mapping.call_action(image_x, image_y)
                    else:
                        log.debug('drag, moves:{}'.format(moves))

                    moves = 0

def main():

    try:
        parser = argparse.ArgumentParser(description="use reMarkable tablet as wireless mouse and keyboard")
        parser.add_argument('--debug', action='store_true', default=False, help="enable debug messages")
        parser.add_argument('--key', type=str, metavar='PATH', help="ssh private key")
        parser.add_argument('--password', default=None, type=str, help="ssh password")
        parser.add_argument('--address', default='10.11.99.1', type=str, help="device address")

        args = parser.parse_args()

        if args.debug:
            print('Debugging enabled...')
            logging.getLogger('').setLevel(logging.DEBUG)
            log.setLevel(logging.DEBUG)

        read_tablet(args)
    except KeyboardInterrupt:
        pass

if __name__ == '__main__':
    main()
